from enum import Enum, unique


CSI = "\033["
RESET = CSI + "0m"


@unique
class AnsiStyle(Enum):
    BOLD = 1
    DIM = 2
    ITALIC = 3
    UNDERLINE = 4
    BLINK = 5
    INVERT = 7
    STRIKE = 9


class AnsiForeColor(Enum):
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    LIGHT_BLACK = 90
    LIGHT_RED = 91
    LIGHT_GREEN = 92
    LIGHT_YELLOW = 93
    LIGHT_BLUE = 94
    LIGHT_MAGENTA = 95
    LIGHT_CYAN = 96
    LIGHT_WHITE = 97
    LBLACK = 90
    LRED = 91
    LGREEN = 92
    LYELLOW = 93
    LBLUE = 94
    LMAGENTA = 95
    LCYAN = 96
    LWHITE = 97


class AnsiBackColor(Enum):
    BLACK = 40
    RED = 41
    GREEN = 42
    YELLOW = 43
    BLUE = 44
    MAGENTA = 45
    CYAN = 46
    WHITE = 47
    LIGHT_BLACK = 100
    LIGHT_RED = 101
    LIGHT_GREEN = 102
    LIGHT_YELLOW = 103
    LIGHT_BLUE = 104
    LIGHT_MAGENTA = 105
    LIGHT_CYAN = 106
    LIGHT_WHITE = 107
    LBLACK = 100
    LRED = 101
    LGREEN = 102
    LYELLOW = 103
    LBLUE = 104
    LMAGENTA = 105
    LCYAN = 106
    LWHITE = 107



import re
from enum import Enum
from typing import List, Optional, Sequence, Tuple, Type, Union, cast  # noqa

from ._const import CSI, RESET, AnsiBackColor, AnsiForeColor, AnsiStyle


_regexp_color_code = re.compile(
    "^#?(?P<red>[0-9a-f]{2})(?P<green>[0-9a-f]{2})(?P<blue>[0-9a-f]{2})$", re.IGNORECASE
)

RGBTuple = Tuple[int, int, int]


class Color:
    def __init__(self, color: Union[str, RGBTuple]) -> None:
        if isinstance(color, str):
            self.__from_color_code(color)
            return

        self.red, self.green, self.blue = color
        self.__validate_color_value(self.red)
        self.__validate_color_value(self.green)
        self.__validate_color_value(self.blue)

    def __from_color_code(self, color_code: str) -> None:
        match = _regexp_color_code.search(color_code)
        if match is None:
            raise ValueError("invalid color code found: {}".format(color_code))

        red, green, blue = match.group(1, 2, 3)
        self.red = int(red, 16)
        self.green = int(green, 16)
        self.blue = int(blue, 16)

    def __validate_color_value(self, n: int) -> None:
        if not (0 <= n <= 255):
            raise ValueError("value must be within 0-255")

    @property
    def color_code(self) -> str:
        return "#{:02x}{:02x}{:02x}".format(self.red, self.green, self.blue)

    def calc_complementary(self):
        rgb = (self.red, self.green, self.blue)
        n = max(rgb) + min(rgb)
        return Color((n - self.red, n - self.green, n - self.blue))


def _normalize_enum(value, enum_class: Type[Enum]):
    if isinstance(value, enum_class):
        return value

    try:
        return enum_class[value.strip().upper()]
    except AttributeError:
        raise TypeError("value must be a {} or a str: actual={}".format(enum_class, type(value)))
    except KeyError:
        raise ValueError(
            "invalid valid found: expected={}, actual={}".format(
                "/".join([item.name for item in enum_class]), value
            )
        )


def _ansi_escape(value: str) -> str:
    return "{}{}m".format(CSI, value)


def _to_ansi_style(style: Union[str, AnsiStyle]) -> str:
    return _ansi_escape("{}".format(_normalize_enum(style, AnsiStyle).value))


def _to_ansi_fg_truecolor(color: Color) -> str:
    return _ansi_escape("38;2;{};{};{}".format(color.red, color.green, color.blue))


def _to_ansi_bg_truecolor(color: Color) -> str:
    return _ansi_escape("48;2;{};{};{}".format(color.red, color.green, color.blue))


def _to_ansi_fg_color(color_name: Union[str, AnsiForeColor]) -> str:
    return _ansi_escape("{}".format(_normalize_enum(color_name, AnsiForeColor).value))


def _to_ansi_bg_color(color_name: Union[str, AnsiBackColor]) -> str:
    return _ansi_escape("{}".format(_normalize_enum(color_name, AnsiBackColor).value))


def _make_ansi_fg_truecolor(color: Union[Color, str, RGBTuple, AnsiForeColor, None]) -> str:
    if isinstance(color, AnsiForeColor):
        return _to_ansi_fg_color(color)

    if isinstance(color, str):
        try:
            c = Color(color)
        except ValueError:
            if color[0] == "#":
                raise

            return _to_ansi_fg_color(color)
    elif isinstance(color, (tuple, list)):
        c = Color(color)
    else:
        c = color  # type: ignore

    return _to_ansi_fg_truecolor(c) if c else ""


def _make_ansi_bg_truecolor(color: Union[Color, str, RGBTuple, AnsiBackColor, None]) -> str:
    if isinstance(color, AnsiBackColor):
        return _to_ansi_bg_color(color)

    if isinstance(color, str):
        try:
            c = Color(color)
        except ValueError:
            if color[0] == "#":
                raise

            return _to_ansi_bg_color(color)
    elif isinstance(color, (tuple, list)):
        c = Color(color)
    else:
        c = color  # type: ignore

    return _to_ansi_bg_truecolor(c) if c else ""


def tcolor(
    string: str,
    color: Union[Color, str, RGBTuple, AnsiForeColor, None] = None,
    bg_color: Union[Color, str, RGBTuple, AnsiBackColor, None] = None,
    styles: Optional[Sequence[Union[str, AnsiStyle]]] = None,
) -> str:
    ansi_fg_color = _make_ansi_fg_truecolor(color)
    ansi_bg_color = _make_ansi_bg_truecolor(bg_color)

    ansi_styles = []  # type: List[str]
    if styles:
        ansi_styles = [_to_ansi_style(style) for style in styles]

    reset = RESET
    if not ansi_fg_color and not ansi_bg_color and not ansi_styles:
        reset = ""

    return "".join(ansi_styles + [ansi_bg_color, ansi_fg_color, string, reset])



import os.path
from typing import Dict

import setuptools


MODULE_NAME = "tcolorpy"
REPOSITORY_URL = "https://github.com/thombashi/{:s}".format(MODULE_NAME)
REQUIREMENT_DIR = "requirements"

pkg_info = {}  # type: Dict[str, str]


def get_release_command_class() -> Dict[str, setuptools.Command]:
    try:
        from releasecmd import ReleaseCommand
    except ImportError:
        return {}

    return {"release": ReleaseCommand}


with open(os.path.join(MODULE_NAME, "__version__.py")) as f:
    exec(f.read(), pkg_info)

with open("README.rst", encoding="utf8") as f:
    LONG_DESCRIPTION = f.read()

with open(os.path.join(REQUIREMENT_DIR, "test_requirements.txt")) as f:
    TESTS_REQUIRES = [line.strip() for line in f if line.strip()]

setuptools.setup(
    name=MODULE_NAME,
    version=pkg_info["__version__"],
    url=REPOSITORY_URL,
    author=pkg_info["__author__"],
    author_email=pkg_info["__email__"],
    description="tcolopy is a Python library to apply true color for terminal text.",
    include_package_data=True,
    keywords=["truecolor", "ansi"],
    license=pkg_info["__license__"],
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    packages=setuptools.find_packages(exclude=["test*"]),
    package_data={MODULE_NAME: ["py.typed"]},
    project_urls={"Source": REPOSITORY_URL, "Tracker": "{:s}/issues".format(REPOSITORY_URL),},
    python_requires=">=3.5",
    extras_require={"test": TESTS_REQUIRES},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Terminals",
        "Topic :: Text Processing",
    ],
    zip_safe=False,
    cmdclass=get_release_command_class(),
)



import pytest

from tcolorpy import Color, tcolor


class Test_truecolor_color:
    @pytest.mark.parametrize(
        ["string", "color", "expected"],
        [
            ["test", "#ff8822", "\x1b[38;2;255;136;34mtest\x1b[0m"],
            ["test", "ff8822", "\x1b[38;2;255;136;34mtest\x1b[0m"],
            ["test", "red", "\x1b[31mtest\x1b[0m"],
            ["test", (255, 136, 34), "\x1b[38;2;255;136;34mtest\x1b[0m"],
            ["test", Color("#ff8822"), "\x1b[38;2;255;136;34mtest\x1b[0m"],
            ["test", Color((255, 136, 34)), "\x1b[38;2;255;136;34mtest\x1b[0m"],
            ["test", None, "test"],
        ],
    )
    def test_normal(self, string, color, expected):
        assert tcolor(string, color=color) == expected

    @pytest.mark.parametrize(
        ["value", "expected"], [["#fffff", ValueError], ["#GGGGGG", ValueError],],
    )
    def test_exception(self, value, expected):
        with pytest.raises(expected):
            tcolor("test", color=value) == expected


class Test_truecolor_bg_color:
    @pytest.mark.parametrize(
        ["string", "color", "expected"],
        [
            ["test", "#ff8822", "\x1b[48;2;255;136;34mtest\x1b[0m"],
            ["test", "ff8822", "\x1b[48;2;255;136;34mtest\x1b[0m"],
            ["test", (255, 136, 34), "\x1b[48;2;255;136;34mtest\x1b[0m"],
            ["test", Color("#ff8822"), "\x1b[48;2;255;136;34mtest\x1b[0m"],
            ["test", Color((255, 136, 34)), "\x1b[48;2;255;136;34mtest\x1b[0m"],
            ["test", None, "test"],
        ],
    )
    def test_normal_bg_color(self, string, color, expected):
        assert tcolor(string, bg_color=color) == expected


class Test_truecolor_fg_bg_color:
    @pytest.mark.parametrize(
        ["string", "color", "bg_color", "expected"],
        [
            ["test", "#111111", "#ffffff", "\x1b[48;2;255;255;255m\x1b[38;2;17;17;17mtest\x1b[0m"],
            ["test", None, None, "test"],
        ],
    )
    def test_normal_fg_bg_color(self, string, color, bg_color, expected):
        assert tcolor(string, color=color, bg_color=bg_color) == expected


class Test_truecolor_styles:
    @pytest.mark.parametrize(
        ["string", "styles", "expected"],
        [
            ["test", [], "test"],
            ["test", ["bold"], "\x1b[1mtest\x1b[0m"],
            ["test", ["dim"], "\x1b[2mtest\x1b[0m"],
            ["test", ["italic"], "\x1b[3mtest\x1b[0m"],
            ["test", ["underline"], "\x1b[4mtest\x1b[0m"],
            ["test", ["blink"], "\x1b[5mtest\x1b[0m"],
            ["test", ["invert"], "\x1b[7mtest\x1b[0m"],
            ["test", ["strike"], "\x1b[9mtest\x1b[0m"],
            ["test", ["bold", "dim"], "\x1b[1m\x1b[2mtest\x1b[0m"],
        ],
    )
    def test_normal_bg_color(self, string, styles, expected):
        assert tcolor("test", styles=styles) == expected
